
from pathlib import Path
import json
import sys

import click

from enviroMS.singleMzSearch import run_molecular_formula_search
from enviroMS.diWorkflow import DiWorkflowParameters, generate_database, run_di_mpi, run_direct_infusion_workflow, run_wdl_direct_infusion_workflow
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulas
from corems.encapsulation.output.parameter_to_json import dump_ms_settings_json, dump_all_settings_json

class Config:
    def __init__(self):
        self.verbose = False
        
pass_config = click.make_pass_decorator(Config, ensure=True)

@click.group()
@click.option('--verbose', is_flag=True, help='print out the results')
@pass_config
def cli(config, verbose):
    
    config.verbose = verbose

@cli.command()
@click.argument('mz', required=True, type=float, )
@click.argument('corems_parameters_filepath', required=True, type=click.Path())
@click.argument('out', required=False, type=click.File('w'), default='-')
@click.option('-e', '--error', 'ppm_error', default=1.0,  help='the marging of mass error (ppm)')
@click.option('-r', '--radical', 'isRadical', default=True, type=bool, help='include radical ion type') 
@click.option('-p', '--protonated', 'isProtonated', default=True, type=bool, help='include (de)-protonated ion type')
@click.option('-a', '--adduct', 'isAdduct', default=False, type=bool, help='include adduct ion type')

@pass_config
def run_search_formula(config, mz, ppm_error, isRadical, isProtonated, isAdduct,out, corems_parameters_filepath):#settings_filepath
    
    '''Search for molecular formula candidates to a given m/z value \n
       corems_parameters_filepath =' CoreMS Parameters File (JSON)' 
       MZ = m/z value FLOAT\n
       out = filename to store results TEXT\n
    '''
    
    #if config.verbose:
    click.echo('', file=out)
    #dump_search_settings_yaml()    
    
    click.echo('Searching formulas for %.5f' % mz, file=out)
    
    click.echo('', file=out)
    
    click.echo('Loading Searching Settings from %s' % corems_parameters_filepath, file=out)
    
    click.echo('',file=out)

    run_molecular_formula_search(mz, out, corems_parameters_filepath)

@cli.command()
@click.argument('corems_parameters_file', required=True, type=str)
@click.option('--jobs','-j', default=4, help="'cpu's'")
def create_database(corems_parameters_file, jobs):
    '''corems_parameters_file: Path for CoreMS JSON Parameters file\n
       jobs: Number of processes to run\n
       "postgresql://postgres:labthomson0102@172.22.113.27:5432/",
    '''
    generate_database(corems_parameters_file, jobs)

@cli.command()
@click.argument('file_paths', required=True, type=str)
@click.argument('output_directory', required=True, type=str)
@click.argument('output_type', required=True, type=str)
@click.argument('corems_json_path', required=True, type=str)
@click.argument('polarity', required=True, type=str)
@click.argument('raw_file_start_scan', required=True, type=int)
@click.argument('raw_file_final_scan', required=True, type=int)
@click.argument('is_centroid', required=True, type=bool)
@click.argument('calibration_ref_file_path', required=False, type=str)
@click.option('--calibrate','-c', default=True)
@click.option('--plot_mz_error', '-e', default=True)
@click.option('--plot_ms_assigned_unassigned','-a', default=True)
@click.option('--plot_c_dbe', '-cb', default=True)
@click.option('--plot_van_krevelen', '-vk', default=True)
@click.option('--plot_ms_classes', '-mc', default=True)
@click.option('--plot_mz_error_classes', '-ec',  default=True)
@click.option('--jobs','-j', default=4, help="'cpu's'")
def run_di_wdl(*args, **kwargs):
    '''Run the Direct Infusion Workflow using wdl'''
    
    run_wdl_direct_infusion_workflow(*args, **kwargs)


@cli.command()
@click.argument('di_workflow_paramaters_file', required=True, type=str)
@click.option('--jobs','-j', default=4, help="'cpu's'")
@click.option('--replicas','-r', default=1, help="data replicas")
@click.option('--tasks','-t', default=4, help="mpi tasks")
@click.option('--mpi','-m', is_flag=True, help="run mpi version")
def run_di(di_workflow_paramaters_file, jobs, replicas, tasks, mpi):
    '''Run the Direct Infusion Workflow\n
       workflow_paramaters_file = json file with workflow parameters\n
       output_types = csv, excel, pandas, json set on the parameter file\n
       corems_json_path = json file with corems parameters\n
       --jobs = number of processes to run in parallel\n 
       --mpi = run on hpc, if omitted will run python's multiprocessing and will duplicate runs on nodes\n
    '''
    if mpi:
        
        run_di_mpi(di_workflow_paramaters_file, tasks, replicas)

    else:    

        run_direct_infusion_workflow(di_workflow_paramaters_file, jobs, replicas)

@cli.command()
@click.argument('lcms_workflow_paramaters_file', required=True, type=str)
@click.option('--jobs','-j', default=4, help="'cpu's'")
@pass_config
def run_lcms(workflow_paramaters_file, jobs):
    #implement a mz search inside the mass spectrum, then run a search for molecular formula and the isotopologues
    pass       

@cli.command()
@click.argument('json_file_name', required=True, type=click.Path())
def dump_corems_template(json_file_name):
    '''Dumps a CoreMS json file template
        to be used as the workflow parameters input 
    '''
    path_obj = Path(json_file_name).with_suffix('.json')
    dump_all_settings_json(file_path=path_obj)

@cli.command()
@click.argument('json_file_name', required=True, type=click.Path())
def dump_corems_enviroms_template(json_file_name):
    '''Dumps a CoreMS json file template
        to be used as the workflow parameters input 
    '''
    path_obj = Path(json_file_name).with_suffix('.json')
    dump_ms_settings_json(file_path=path_obj)

@cli.command()
@click.argument('json_file_name', required=True, type=click.Path())
def dump_di_template(json_file_name):
    '''Dumps a json file template
        to be used as the workflow parameters input
    '''
    ref_lib_path = Path(json_file_name).with_suffix('.json')
    with open(ref_lib_path, 'w') as workflow_param:
    
        json.dump(DiWorkflowParameters().__dict__, workflow_param, indent=4)



import click
from tabulate import tabulate
from pathlib import Path
from corems.mass_spectrum.input.numpyArray import ms_from_array_centroid
from corems.encapsulation.input import parameter_from_json
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulas


def run_molecular_formula_search(mz, out, parameters_filepath):
    
    mz = [mz]
    abundance = [1]
    rp, s2n = [[1],[1]]
    dataname = Path(str(out))
    
    mass_spectrum_obj = ms_from_array_centroid(mz, abundance, rp, s2n, dataname)
    
    parameter_from_json.load_and_set_parameters_ms(mass_spectrum_obj, parameters_path=parameters_filepath)

    mass_spectrum_obj.molecular_search_settings.use_min_peaks_filter = False
    mass_spectrum_obj.molecular_search_settings.use_min_peaks_filter = 10
    mass_spectrum_obj.molecular_search_settings.use_isotopologue_filter = False

    click.echo('Searching for molecular formulas within %.3f and %.3f ppm' % (mass_spectrum_obj.molecular_search_settings.min_ppm_error, mass_spectrum_obj.molecular_search_settings.max_ppm_error))

    SearchMolecularFormulas(mass_spectrum_obj, find_isotopologues=True).run_worker_ms_peaks([mass_spectrum_obj[0]])
    
    ms_peak = mass_spectrum_obj[0]
    
    if ms_peak:
        
        header = ['Molecular Formula',  'Calculated m/z', 'Mass Error', 'DBE', 'Ion Type']
        
        results = []
        
        for formula in ms_peak:
            
            results.append([formula.to_string, formula.mz_calc, formula.mz_error, formula.dbe, formula.ion_type])
              
        
        click.echo(tabulate(results, headers=header, floatfmt=("s", ".5f", ".5f", ".1f", "s"  )), file=out)
        click.echo('', file=out)        
        
    else:        
        
        click.echo("Could not find a possible molecular formula match for the m/z %.5f" % mz[0], file=out)
        click.echo('', file=out)


__version__ = "4.2.1"
__doc__ = '''
# Table of Contents  
- Introduction
  - [EnviroMS](#EnviroMS)  
  - [Version](#current-version)  
  - [Data Input](#data-input-formats)  
  - [Data Output](#data-output-formats)  
  - [Data Structure](#data-structure-types)  
  - [Features](#molecular-formulae-search-and-assignment)  
- Installation
  - [PyPi](#enviroms-installation)  
- Execution:  
  - [CLI](#running-the-workflow)  
  - [MiniWDL](#MiniWDL)  
  - [Docker Container](#enviroms-docker)  
  

# EnviroMS

**EnviroMS** is a workflow for natural organic matter data processing and annotation

## Current Version

### `4.1.1`

### Data input formats

- Generic mass list in profile and centroid mode (include all delimiters types and Excel formats)

### Data output formats

- Pandas data frame (can be saved using pickle, h5, etc)
- Text Files (.csv, tab separated .txt, etc)
- Microsoft Excel (xlsx)
- Automatic JSON for workflow metadata
- Self-containing Hierarchical Data Format (.hdf5) including raw data and ime-series data-point for processed data-sets with all associated workflow metadata (JSON)

### Data structure types

- FT-ICR MS
- LC-FT-ICR MS

### Molecular formulae search and assignment

- Automatic local (SQLite) or external (PostgreSQL) database check, generation, and search
- Automatic molecular formulae assignments algorithm for ESI(-) MS for natural organic matter analysis
- Automatic fine isotopic structure calculation and search for all isotopes
- Flexible Kendrick normalization base
- Kendrick filter using density-based clustering
- Kendrick classification
- Hetero atoms classification and visualization


## EnviroMS Installation

- PyPi:     
```bash
pip3 install enviroms
```

- From source:
 ```bash
pip3 install --editable .
```

To be able to open thermo raw files a installation of pythonnet is needed:
- Windows: 
    ```bash
    pip3 install pythonnet
    ```

- Mac and Linux:
    ```bash
    brew install mono
    pip3 install pythonnet   
    ```

## Running the workflow


```bash
enviroMS dump-corems-enviroms-template EnviromsFile.json
```
```bash
enviroMS dump-corems-template CoremsFile.json
```

 Modify the EnviromsFile.json and CoremsFile.json accordingly to your dataset and workflow parameters
make sure to include CoremsFile.json path inside the EnviromsFile.json: "corems_json_path": "path_to_CoremsFile.json" 

```bash
enviroMS run-di path_to_MetamsFile.json
```

## MiniWDL 
- Change wdl/enviroms_input.json to specify the data location

- Change data/CoremsFile.json to specify the workflow parameters

Install miniWDL:
```bash
pip3 install miniwdl
```

Call:
```bash
miniwdl run wdl/enviroMS.wdl -i wdl/enviroms_input.json --verbose --no-cache --copy-input-files
```

WARNING ** Current mode only allows for multiprocessing in a single node and it defaults to one job at a time. 
To use multiprocessing mode modify the parameter "runDirectInfusion.jobs_count" in the enviroMS.wdl and modify the parameter "MolecularFormulaSearch.url_database" on CoremsFile.json to point to a Postgresql url. The default is set to use SQLite and it will fail on multiprocessing mode.

## EnviroMS Docker 

A docker image containing the EnviroMS command line as code entry-point

If you don't have docker installed, the easiest way is to [install docker for desktop](https://hub.docker.com/?overlay=onboarding)

- Pull from Docker Registry:

    ```bash
    docker pull corilo/enviroms:latest
    
    ```

- Or to build the image from source:

    ```bash
    docker build -t enviroms:latest .
    ```
- Run Workflow from Container:

    $(data_dir) = dir_containing the FT-ICR MS data, EnviromsFile.json and CoremsFile.json
    
    ```bash
    docker run -v $(data_dir):/enviroms/data corilo/enviroms:latest enviroMS run-di /enviroms/data/EnviromsFile.json    
    ```

- Save a new parameters file template:
    
    ```bash
    docker run -v $(data_dir):/enviroms/data corilo/enviroms:latest enviroMS dump_json_template /enviroms/data/EnviromsFile.json    
    ```
    
    ```bash
    docker run -v $(data_dir):/metaB/data corilo/enviroms:latest enviroMS dump_corems_json_template /metaB/data/CoremsFile.json
    ```

## Disclaimer

This material was prepared as an account of work sponsored by an agency of the
United States Government.  Neither the United States Government nor the United
States Department of Energy, nor Battelle, nor any of their employees, nor any
jurisdiction or organization that has cooperated in the development of these
materials, makes any warranty, express or implied, or assumes any legal
liability or responsibility for the accuracy, completeness, or usefulness or
any information, apparatus, product, software, or process disclosed, or
represents that its use would not infringe privately owned rights.

Reference herein to any specific commercial product, process, or service by
trade name, trademark, manufacturer, or otherwise does not necessarily
constitute or imply its endorsement, recommendation, or favoring by the United
States Government or any agency thereof, or Battelle Memorial Institute. The
views and opinions of authors expressed herein do not necessarily state or
reflect those of the United States Government or any agency thereof.

                 PACIFIC NORTHWEST NATIONAL LABORATORY
                              operated by
                                BATTELLE
                                for the
                   UNITED STATES DEPARTMENT OF ENERGY
                    under Contract DE-AC05-76RL01830
'''


import cProfile
from dataclasses import dataclass, asdict
import json
from multiprocessing import Pool
import os
from pathlib import Path

import click
from tqdm import tqdm

from matplotlib import pyplot as plt
from corems.mass_spectrum.calc.Calibration import MzDomainCalibration
from corems.molecular_id.factory.classification import HeteroatomsClassification
from corems.mass_spectrum.input.massList import ReadMassList

from corems.molecular_id.search.priorityAssignment import OxygenPriorityAssignment
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulas
from corems.transient.input.brukerSolarix import ReadBrukerSolarix

from corems.molecular_id.factory.MolecularLookupTable import MolecularCombinations
from corems.encapsulation.factory.processingSetting import MolecularFormulaSearchSettings
from corems.encapsulation.input.parameter_from_json import load_and_set_parameters_class

@dataclass
class DiWorkflowParameters:

    # input type: masslist, bruker_transient, thermo_reduced_profile
    
    # scans to sum for thermo raw data, reduce profile
    raw_file_start_scan: int = 1
    raw_file_final_scan: int = 7

    # input output paths
    file_paths: tuple = ('data/...', 'data/...')
    output_directory: str = 'data/...'
    output_group_name: str = '...'
    output_type: str = 'csv'

    # polarity for masslist input
    polarity: int = -1
    is_centroid:bool = False
    # corems settings
    corems_json_path: str = 'data/CoremsFile.json'

    # calibration
    calibrate: bool = True
    calibration_ref_file_path: str = 'data/SRFA.ref'

    # plots
    plot_mz_error: bool = True
    plot_ms_assigned_unassigned: bool = True

    plot_c_dbe: bool = True
    plot_van_krevelen: bool = True
    plot_ms_classes: bool = True
    plot_mz_error_classes: bool = True

    def to_json(self):
        return json.dumps(asdict(self))

def run_thermo_reduce_profile(file_location, corems_params_path, first_scan, last_scan):

    from corems.mass_spectra.input import rawFileReader
    parser =  rawFileReader.ImportMassSpectraThermoMSFileReader(file_location)
    
    mass_spectrum = parser.get_average_mass_spectrum_in_scan_range(first_scan=1, last_scan=5)
    return mass_spectrum

def run_bruker_transient(file_location, corems_params_path):

    with ReadBrukerSolarix(file_location) as transient:

        transient.set_parameter_from_json(corems_params_path) 
        mass_spectrum = transient.get_mass_spectrum(plot_result=False, auto_process=True)
    
    return mass_spectrum

def get_masslist(file_location, corems_params_path, polarity, is_centroid):

    if is_centroid:
        reader = ReadMassList(file_location)
    else:
       reader = ReadMassList(file_location, header_lines=7, isCentroid=False, isThermoProfile=True)
    
    reader.set_parameter_from_json(parameters_path=corems_params_path)

    return(reader.get_mass_spectrum(polarity=polarity))

def run_assignment(file_location, workflow_params):

    file_path = Path(file_location)
    
    if file_path.suffix == '.raw':
    
        first_scan, last_scan = workflow_params.raw_file_start_scan, workflow_params.raw_file_final_scan    
        mass_spectrum = run_thermo_reduce_profile(file_location, workflow_params, first_scan, last_scan)

    elif file_path.suffix == '.d':

        mass_spectrum = run_bruker_transient(file_location, workflow_params.corems_json_path)

    
    elif file_path.suffix == '.txt' or file_path.suffix == 'csv' or file_path.suffix == '.tsv':
        
        mass_spectrum = get_masslist(file_location, workflow_params.corems_json_path, 
                        polarity=workflow_params.polarity, is_centroid=workflow_params.is_centroid)


    mass_spectrum.set_parameter_from_json(workflow_params.corems_json_path)
    
    if workflow_params.calibrate:

        ref_file_location = Path(workflow_params.calibration_ref_file_path)

        MzDomainCalibration(mass_spectrum, ref_file_location).run()

    # force it to one job. daemon child can not have child process 
    mass_spectrum.molecular_search_settings.db_jobs = 1

    SearchMolecularFormulas(mass_spectrum, first_hit=False).run_worker_mass_spectrum()

    
    return mass_spectrum

def generate_database(corems_parameters_file, jobs):

    '''corems_parameters_file: Path for CoreMS JSON Parameters file
       --jobs: Number of processes to run   
    '''
    click.echo('Loading Searching Settings from %s' % corems_parameters_file)

    molecular_search_settings = load_and_set_parameters_class('MolecularSearch', MolecularFormulaSearchSettings(), parameters_path=corems_parameters_file)
    molecular_search_settings.db_jobs = jobs
    molecular_search_settings.url_database = None
    MolecularCombinations().runworker(molecular_search_settings)

def read_workflow_parameter(di_workflow_paramaters_json_file):

    with open(di_workflow_paramaters_json_file, 'r') as infile:
        return DiWorkflowParameters(**json.load(infile))

def create_plots(mass_spectrum, workflow_params, dirloc):

    ms_by_classes = HeteroatomsClassification(mass_spectrum, choose_molecular_formula=False)

    if workflow_params.plot_ms_assigned_unassigned:
        print("Plotting assigned vs. unassigned mass spectrum")
        ax_ms = ms_by_classes.plot_ms_assigned_unassigned()
        plt.savefig(dirloc / "assigned_unassigned.png", bbox_inches='tight')
        plt.clf()

    if workflow_params.plot_mz_error:
        print("Plotting mz_error")
        ax_ms = ms_by_classes.plot_mz_error()
        plt.savefig(dirloc / "mz_error.png", bbox_inches='tight')
        plt.clf()

    if workflow_params.plot_van_krevelen:
        van_krevelen_dirloc = dirloc / "van_krevelen"
        van_krevelen_dirloc.mkdir(exist_ok=True, parents=True)  

    if workflow_params.plot_c_dbe:
        c_dbe_dirloc = dirloc / "dbe_vs_c"
        c_dbe_dirloc.mkdir(exist_ok=True, parents=True)      

    if workflow_params.plot_ms_classes:
        ms_class_dirloc = dirloc / "ms_class"
        ms_class_dirloc.mkdir(exist_ok=True, parents=True)  

    if workflow_params.plot_mz_error_classes:
        mz_error_class_dirloc = dirloc / "mz_error_class"
        mz_error_class_dirloc.mkdir(exist_ok=True, parents=True)  

    pbar = tqdm(ms_by_classes.get_classes())

    for classe in pbar:

        pbar.set_description_str(desc="Plotting results for class {}".format(classe), refresh=True)

        if workflow_params.plot_van_krevelen:
            ax_c = ms_by_classes.plot_van_krevelen(classe)
            plt.savefig(van_krevelen_dirloc / "{}.png".format(classe), bbox_inches='tight')
            plt.clf()

        if workflow_params.plot_mz_error_classes:
            ax_c = ms_by_classes.plot_mz_error_class(classe)
            plt.savefig(mz_error_class_dirloc / "{}.png".format(classe), bbox_inches='tight')
            plt.clf()

        if workflow_params.plot_ms_classes:
            ax_c = ms_by_classes.plot_ms_class(classe)
            plt.savefig(ms_class_dirloc / "{}.png".format(classe), bbox_inches='tight')
            plt.clf()

        if workflow_params.plot_c_dbe:
            ax_c = ms_by_classes.plot_dbe_vs_carbon_number(classe)
            plt.savefig(c_dbe_dirloc / "{}.png".format(classe), bbox_inches='tight')
            plt.clf()

def workflow_worker(args):

    file_location, workflow_params_json_str = args

    workflow_params = DiWorkflowParameters(**json.loads(workflow_params_json_str))
    
    mass_spec = run_assignment(file_location, workflow_params)

    dirloc = Path(workflow_params.output_directory) / mass_spec.sample_name

    dirloc.mkdir(exist_ok=True, parents=True)

    output_path = dirloc / mass_spec.sample_name

    eval('mass_spec.to_{OUT_TYPE}(output_path)'.format(OUT_TYPE=workflow_params.output_type))

    create_plots(mass_spec, workflow_params, dirloc)

    return 'Success' + str(os.getpid())

def cprofile_worker(file_location, workflow_params_json_str):

    cProfile.runctx('run_assignment(file_location, workflow_params)', globals(), locals(), 'di-fticr-di.prof')
    # stats = pstats.Stats("topics.prof")
    # stats.strip_dirs().sort_stats("time").print_stats() 

def run_wdl_direct_infusion_workflow(*args, **kwargs):
    
    cores = kwargs.get('jobs')
    del kwargs['jobs']
    kwargs["polarity"] = -1 if kwargs.get('polarity') == 'negative' else 1
    
    workflow_params = DiWorkflowParameters(**kwargs)
    workflow_params.file_paths = workflow_params.file_paths.split(",")
    # workflow_params.output_directory = kwargs.get['output_directory']
    # workflow_params.output_type = kwargs.get['output_type']
    # workflow_params.corems_json_path = kwargs.get['corems_json_path']
    # workflow_params.polarity = -1 if kwargs.get['polarity'] == 'negative' else 1
    # workflow_params.raw_file_start_scan = kwargs.get['raw_file_start_scan']
    # workflow_params.raw_file_final_scan = kwargs.get['raw_file_final_scan']
    # workflow_params.is_centroid = kwargs.get['is_centroid']
    # workflow_params.calibration_ref_file_path = kwargs.get['calibration_ref_file_path']

    # workflow_params.calibrate = kwargs.get['calibrate']
    # workflow_params.calibrate = kwargs.get['plot_mz_error']
    # workflow_params.calibrate = kwargs.get['plot_ms_assigned_unassigned']
    # workflow_params.calibrate = kwargs.get['plot_c_dbe']
    # workflow_params.calibrate = kwargs.get['plot_van_krevelen']
    # workflow_params.calibrate = kwargs.get['plot_ms_classes']
    # workflow_params.calibrate = kwargs.get['plot_mz_error_classes']
    # workflow_params.calibrate = kwargs.get['calibrate']

    dirloc = Path(workflow_params.output_directory)
    dirloc.mkdir(exist_ok=True)

    pool = Pool(cores)

    worker_args = [(file_path, workflow_params.to_json()) for file_path in workflow_params.file_paths]
    file_path = Path(worker_args[0][0])
    #for worker_arg in worker_args:
    #    workflow_worker(worker_arg)
    
    for i, results in enumerate(pool.imap_unordered(workflow_worker, worker_args), 1):
        pass
    pool.close()
    pool.join()

def run_direct_infusion_workflow(workflow_params_file, jobs, replicas):

    click.echo('Loading Searching Settings from %s' % workflow_params_file)

    workflow_params = read_workflow_parameter(workflow_params_file)

    dirloc = Path(workflow_params.output_directory)
    dirloc.mkdir(exist_ok=True)

    worker_args = replicas * [(file_path, workflow_params.to_json()) for file_path in workflow_params.file_paths]

    cores = jobs
    pool = Pool(cores)

    for worker_arg in worker_args:
        workflow_worker(worker_arg)
    # for i, results in enumerate(pool.imap_unordered(workflow_worker, worker_args), 1):

    #    pass

    pool.close()
    pool.join()

def run_di_mpi(workflow_params_file, tasks, replicas):

    import os, sys
    from mpi4py import MPI
    # from mpi4py.futures import MPIPoolExecutor
    sys.path.append(os.getcwd())

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    workflow_params = read_workflow_parameter(workflow_params_file)
    all_worker_args = replicas * [(file_path, workflow_params.to_json()) for file_path in workflow_params.file_paths]

    # worker_args = comm.scatter(all_worker_args, root=0)

    # will only run tasks up to the number of files paths selected in the EnviroMS File
    if len(all_worker_args) <= size:

        workflow_worker(all_worker_args[0])

    else:

        print("Tasks needs to be the same size of the input data count, , until you find time to come and help to code this section :D")
